from __future__ import unicode_literals

from django.db import migrations, models

import django_courier.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0004_template_targets'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='sitecontact',
            managers=[
                ('objects', django_courier.models.SiteContactManager()),
            ],
        ),
        migrations.AddField(
            model_name='sitecontact',
            name='name',
            field=models.CharField(blank=True, max_length=500,
                                   verbose_name='name'),
        ),
    ]



import collections
from typing import Any, List, Mapping, cast

import itertools
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.template import Context
from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from . import settings, templates
from .base import Contact, AbstractContactable, AbstractContactNetworkN
from .backends import get_backends_from_settings, get_backends

__ALL_TARGETS__ = collections.OrderedDict()


def get_model_variables(label, value, cls, depth=0):
    if cls is None:
        yield {'label': label, 'value': value}
        return
    attrs = []
    assert issubclass(cls, models.Model)
    relations = []
    for field in cls._meta.fields:
        if field.is_relation:
            relations.append(field)
        else:
            attrs.append({
                'label': field.verbose_name.title(),
                'value': '{}.{}'.format(value, field.name),
            })
    yield {'label': label, 'value': value, 'attrs': attrs}
    if depth > 1:
        return
    for field in relations:
        sub_value = '{}.{}'.format(value, field.name)
        sub_label = '{} {}'.format(label, field.verbose_name.title())
        for item in get_model_variables(
                sub_label, sub_value, field.related_model, depth+1):
            yield item


class CourierOptions(object):
    """
    Options for Courier extensions
    """

    ALL_OPTIONS = ('notifications',)
    # list of notification code names
    notifications = []

    def __init__(self, meta):
        """
        Set any options provided, replacing the default values
        """
        if meta is not None:
            for key, value in meta.__dict__.items():
                if key in self.ALL_OPTIONS:
                    setattr(self, key, value)
                elif not key.startswith('_'):  # ignore private parts
                    raise ValueError(
                        'CourierMeta has invalid attribute: {}'.format(key))


class CourierModelBase(models.base.ModelBase):
    """
    Metaclass for Courier extensions.

    Deals with notifications on CourierOptions
    """
    def __new__(mcs, name, bases, attributes):
        new = super(CourierModelBase, mcs).__new__(
            mcs, name, bases, attributes)
        meta = attributes.pop('CourierMeta', None)
        setattr(new, '_courier_meta', CourierOptions(meta))
        return new


class CourierModel(models.Model, metaclass=CourierModelBase):
    """
    Base class for models that implement notifications
    """

    class Meta:
        abstract = True

    def issue_notification(self, codename: str,
                           recipients: 'AbstractContactNetworkN' = None,
                           sender: 'AbstractContactNetworkN' = None):
        ct = ContentType.objects.get_for_model(self)
        notification = Notification.objects.get(
            codename=codename, content_type=ct)
        notification.issue(self, recipients, sender)


class NotificationManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, app_label, model, codename):
        return self.get(
            codename=codename,
            content_type=ContentType.objects.db_manager(
                self.db).get_by_natural_key(app_label, model),
        )


class CourierParam:
    REQUIRED_PARAMS = {'codename', 'description'}
    OPTIONAL_PARAMS = {'use_recipient': True, 'use_sender': True,
                       'sender_model': '', 'recipient_model': '',
                       'required': False}

    def __init__(self, codename, description, **kwargs):
        self.params = {
            'codename': codename,
            'description': description,
        }
        for key, default in self.OPTIONAL_PARAMS.items():
            if key in kwargs:
                self.params[key] = kwargs[key]
                del kwargs[key]
            else:
                self.params[key] = default

    def params_equal(self, notification):
        for key in self.params:
            if getattr(notification, key) != self.params[key]:
                return False
        return True

    def set_params(self, notification):
        for key in self.params:
            setattr(notification, key, self.params[key])

    def create(self, content_type):
        new = Notification(content_type=content_type)
        self.set_params(new)
        return new

    @property
    def codename(self):
        return self.params['codename']


class Notification(models.Model):
    """
    Base class for all notifications
    """

    codename = models.CharField(_('codename'), max_length=100)
    content_type = models.ForeignKey(
        to=ContentType, on_delete=models.CASCADE,
        verbose_name=_('content type'))
    description = models.TextField(_('description'))
    use_sender = models.BooleanField(_('use sender'))
    sender_model = models.CharField(
        _('sender model'), blank=True, max_length=500)
    use_recipient = models.BooleanField(_('use recipient'))
    recipient_model = models.CharField(
        _('recipient model'), blank=True, max_length=500)
    required = models.BooleanField(
        _('required'), default=False,
        help_text=_('If true, triggering the notification will throw an '
                    'error if there is no available template/contact'))

    objects = NotificationManager()

    def __str__(self):
        return "{} | {} | {}".format(
            self.content_type.app_label, self.content_type, self.codename)

    def natural_key(self):
        return self.content_type.natural_key() + (self.codename,)

    natural_key.dependencies = ['contenttypes.contenttype']

    def issue(self, content,
              recipients: AbstractContactNetworkN=None,
              sender: AbstractContactNetworkN=None):
        """
        To send a notification to a user, get all the user's active methods.
        Then get the backend for each method and find the relevant template
        to send (and has the said notification). Send that template with
        the parameters with the backend.

        :param content: model object that the notification is about
        :param recipients: either a user, or None if no logical recipient
        :param sender: user who initiated the notification
        :return: None
        """
        # check
        parameters = {
            'content': content,
        }
        networks = {}
        if self.use_recipient and (recipients is not None):
            networks['re'] = recipients
        elif self.use_recipient:
            raise RuntimeError(
                'Model specified "use_recipient" for notification but '
                'recipient missing on issue_notification ')
        elif recipients is not None:
            raise RuntimeError('Recipient added to issue_notification, '
                               'but is not specified in CourierMeta')

        if self.use_sender and (sender is not None):
            networks['se'] = sender
            parameters['sender'] = sender
        elif self.use_sender:
            raise RuntimeError(
                'Model specified "use_sender" for notification but sender '
                'missing on issue_notification ')
        elif sender is not None:
            raise RuntimeError('Sender added to issue_notification, but is '
                               'not specified in CourierMeta')

        contactable_list = {
            'si': (SiteContact.objects,),
        }
        for key, network in networks.items():
            for channel in settings.CHANNELS:
                if channel != '':
                    key = key + ':' + channel
                contactable_list[key] = network.get_contactables(channel)

        contact_set = collections.defaultdict(dict)
        for key, contactables in contactable_list.items():
            for c_able in contactables:
                for contact in c_able.get_contacts_for_notification(self):
                    contact_set[key][contact] = c_able

        sent = False
        for key, contact_dict in contact_set.items():
            if self.send_messages(contact_dict, parameters,
                                  Template.objects.filter(target=key)):
                sent = True
        if not sent and self.required:
            raise RuntimeError('Notification required, but no message sent')

    def send_messages(
            self, contacts: Mapping[Contact, AbstractContactable],
            generic_params: Mapping[str, Any], template_queryset):

        def _get_backend_message(protocol):
            backends = get_backends_from_settings(protocol)
            # now get all the templates for these backends
            for be in backends:
                tpl = template_queryset.filter(
                    backend=be.ID, notification=self, is_active=True).first()
                if tpl is not None:
                    return be, tpl
            return None

        # per-protocol message cache
        cache = {}
        sent = False
        for contact, contactable in contacts.items():
            params = generic_params.copy()
            params['contact'] = contact
            params['recipient'] = contactable
            proto = contact.protocol
            if proto not in cache:
                cache[proto] = _get_backend_message(proto)
            if cache[proto] is not None:
                backend, template = cache[proto]
                message = backend.build_message(
                    template.subject, template.content, params)
                # We're catching all exceptions here because some people
                # are bad people and can't subclass properly
                try:
                    backend.send_message(contact, message)
                    sent = True
                except Exception as e:
                    FailedMessage.objects.create(
                        backend=backend.ID,
                        name=contact.name,
                        address=contact.address,
                        message=str(message),
                        error=str(e),
                    )
        return sent

    def preview(self, backend_id, message):
        backend = [b for b in get_backends() if b.ID == backend_id][0]
        content_model = self.content_type.model_class()
        params = templates.PreviewParameters(
            content_model, self.get_sender_model(), self.get_recipient_model())
        return backend.preview_message('', message, params)

    def get_sender_model(self):
        if self.sender_model:
            return apps.get_model(self.sender_model)
        if settings.SENDER_MODEL:
            return apps.get_model(settings.SENDER_MODEL)
        return None

    def get_recipient_model(self):
        if self.recipient_model:
            return apps.get_model(self.recipient_model)
        if settings.RECIPIENT_MODEL:
            return apps.get_model(settings.RECIPIENT_MODEL)
        return None

    def get_variables(self):
        sender_model = self.get_sender_model()
        recipient_model = self.get_recipient_model()
        content_model = self.content_type.model_class()
        return list(itertools.chain(
            get_model_variables('Content', 'content', content_model),
            get_model_variables('Sender', 'sender', sender_model),
            get_model_variables('Recipient', 'recipient', recipient_model)
        )) + [
            {'label': 'Contact', 'value': 'contact', 'attrs': [
                {'label': 'Name', 'value': 'contact.name'},
                {'label': 'Protocol', 'value': 'contact.protocol'},
                {'label': 'Address', 'value': 'contact.address'},
            ]},
        ]


class Template(models.Model):

    class Meta:
        verbose_name = _('template')

    notification = models.ForeignKey(
        to=Notification, on_delete=models.PROTECT,
        verbose_name=_('notification'))
    backend = models.CharField(_('backend'), max_length=100)
    subject = models.CharField(_('subject'), max_length=500, blank=True)
    content = models.TextField(_('content'))
    target = models.CharField(
        verbose_name=_('target'), max_length=103, default='re',
        help_text=_('Who this message actually gets sent to.'))
    is_active = models.BooleanField(
        _('is active'), default=True,
        help_text=_('When not active, the template will be ignored'))

    objects = models.Manager()

    def render(self, parameters: dict, autoescape=True):
        content = cast(str, self.content)
        template = templates.from_string(content)
        context = Context(parameters, autoescape=autoescape)
        return template.render(context)


class SiteContactManager(models.Manager, AbstractContactable):
    use_in_migrations = True

    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[Contact]:
        for pref in SiteContactPreference.objects.filter(
                notification=notification, is_active=True):
            sc = pref.site_contact
            yield Contact(sc.name, sc.protocol, sc.address)


# can't make this subclass AbstractContact or fields become unset-able
class SiteContact(models.Model):

    class Meta:
        verbose_name = _('site contact')
        unique_together = (('address', 'protocol'),)

    name = models.CharField(_('name'), blank=True, max_length=500)
    protocol = models.CharField(_('protocol'), max_length=100)
    address = models.CharField(_('address'), max_length=500)

    objects = SiteContactManager()

    def __str__(self):
        return self.name


class SiteContactPreference(models.Model):

    site_contact = models.ForeignKey(SiteContact, on_delete=models.CASCADE)
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE)
    is_active = models.BooleanField(_('is active'))

    objects = models.Manager()


class FailedMessage(models.Model):

    class Meta:
        verbose_name = _('failed message')

    backend = models.CharField(_('backend'), max_length=100)
    contact_name = models.CharField(_('contact name'), max_length=500)
    address = models.CharField(_('address'), max_length=500)
    message = models.TextField(_('message'))
    error = models.TextField(_('error'))
    created_at = models.DateTimeField(_('created at'), auto_now_add=True)

    objects = models.Manager()

    def __str__(self):
        return '{}:{} @ {}'.format(self.backend.PROTOCOL,
                                   self.address, self.created_at)


def _init_targets():
    """
    Intialize channel settings
    """
    if not __ALL_TARGETS__:
        __ALL_TARGETS__['si'] = _('Site Contacts')
        for key, (se_name, re_name) in settings.CHANNELS.items():
            if key == '':
                __ALL_TARGETS__['re'] = re_name
                __ALL_TARGETS__['se'] = se_name
            else:
                __ALL_TARGETS__['re:' + key] = re_name
                __ALL_TARGETS__['se:' + key] = se_name


def get_targets():
    _init_targets()
    return __ALL_TARGETS__


def get_target_choices():
    _init_targets()
    return __ALL_TARGETS__.items()



"""
URL patterns for testing app.

We don't have any fore now.
"""
from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.staticfiles.views import serve

from . import urls

urlpatterns = [

    url(r'^', include(urls)),
    url('^admin/', admin.site.urls),
] + [
    url(r'^static/(?P<path>.*)$', serve),
]



"""
Django settings for django_courier demo
"""

from tests.settings import *  # noqa: F403,F401

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django_courier',
    'tests.apps.CourierDemoConfig',
]

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'



from django.template import TemplateSyntaxError
from django.test import TestCase

from django_courier import models, templates


class MultipartMessageTests(TestCase):

    @staticmethod
    def test_all():
        mpm_dict = {
            'subject': 'foo',
            'text': 'A message about foo',
            'html': 'A <blink>message</blink> about foo',
        }
        mpm = templates.MultipartMessage.from_dict(mpm_dict)
        new_dict = mpm.to_dict()
        assert new_dict == mpm_dict
        serialized = str(mpm)
        new_mpm = templates.MultipartMessage.from_string(serialized)
        assert new_mpm.subject == mpm.subject
        assert new_mpm.text == mpm.text
        assert new_mpm.html == mpm.html
        # test to mail with both text & html
        mail = mpm.to_mail()
        assert mail.body == mpm_dict['text']
        assert mail.subject == mpm_dict['subject']
        assert len(mail.alternatives) == 1
        content, mime = mail.alternatives[0]
        assert content == mpm_dict['html']
        assert mime == 'text/html'
        # test to mail with text only
        mpm.html = ''
        mail = mpm.to_mail()
        assert mail.content_subtype == 'plain'
        assert mail.body == mpm_dict['text']
        # test to mail with html only
        mpm.html = mpm_dict['html']
        mpm.text = ''
        mail = mpm.to_mail()
        assert mail.content_subtype == 'html'
        assert mail.body == mpm_dict['html']


class TemplateRender(TestCase):

    @staticmethod
    def test_blank():
        assert '' == models.Template(content='').render({})

    @staticmethod
    def test_non_html():
        template = models.Template(content='{{ param }}')
        assert 'A&W' == template.render({'param': 'A&W'}, autoescape=False)

    @staticmethod
    def test_html():
        template = models.Template(content='{{ param }}')
        assert 'A&amp;W' == template.render({'param': 'A&W'})

    def test_syntax_error(self):
        template = models.Template(content='{% extends "%}')
        self.assertRaises(TemplateSyntaxError, lambda: template.render({}))

    @staticmethod
    def test_syntax_bad_format():
        template = models.Template(content='{{ }')
        assert '{{ }' == template.render({'param': 'A&W'})

    @staticmethod
    def test_param_missing():
        template = models.Template(content='{{ param }}')
        assert '' == template.render({})


class EmailRenderTests(TestCase):

    @staticmethod
    def test_no_parts():
        params = {'message': 'There\'s a new burger at A&W'}
        email = "Hi, John Doe, here is a message: {{ message }}"
        text = "Hi, John Doe, here is a message: There's a new burger at A&W"
        result = templates.email_parts('', email, params)
        assert result.subject == ''
        assert result.text == text

    @staticmethod
    def test_basic():
        params = {'message': 'There\'s a new burger at A&W'}
        email = """{% block subject %}Test subject{% endblock %}
        {% block text_body %}
        Hi, John Doe, here is a message: {{ message }}
        {% endblock %}
        {% block html_body %}
        <p>Hi, John Doe, here is a message: {{ message }}</p>
        {% endblock %}
        """
        text = "Hi, John Doe, here is a message: There's a new burger at A&W"
        html = "<p>Hi, John Doe, here is a message: " \
               "There&#39;s a new burger at A&amp;W</p>"
        result = templates.email_parts('', email, params)
        assert result.subject == 'Test subject'
        assert result.text == text
        assert result.html == html


class MarkdownEmailRenderTests(TestCase):

    @staticmethod
    def test_basic():
        params = {
            'firstname': 'Jon',
            'lastname': 'Doe',
            'message': 'There\'s a *new* burger at A&W'
        }
        subject = 'Hey {{ firstname }}'
        body = "Hi {{ firstname }} {{ lastname}},\n\n" \
               "Here is a message: {{ message }}"
        text = "Hi Jon Doe,\n\n" \
               "Here is a message: There's a *new* burger at A&W"
        html = "<p>Hi Jon Doe,</p>\n\n" \
               "<p>Here is a message: There&#8217;s a *new* burger " \
               "at A&amp;W</p>\n"
        result = templates.email_md(subject, body, params)
        assert result.subject == 'Hey Jon'
        assert result.text == text
        assert result.html == html



"""
Creates permissions for all installed apps that need permissions.
"""
from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, router


class Command(BaseCommand):
    args = ''
    help = 'Creates notifications based on CourierMeta instances in classes'

    def handle(self, *args, **options):
        for app in apps.get_app_configs():
            make_notifications(app)


def make_notifications(app_config, verbosity=2, dry_run=False,
                       using=DEFAULT_DB_ALIAS):
    if not app_config.models_module:
        return

    app_label = app_config.label
    try:
        app_config = apps.get_app_config(app_label)
        contenttype_class = apps.get_model('contenttypes', 'ContentType')
        notification_class = apps.get_model('django_courier', 'Notification')
    except LookupError:
        return

    if not router.allow_migrate_model(using, notification_class):
        return

    # This will hold the notifications we're looking for as
    # (content_type, codename)
    searched_notifications = []
    # The code names and content types that should exist.
    content_types = set()
    for cls in app_config.get_models():
        meta = getattr(cls, '_courier_meta', None)
        if meta is not None:
            # Force looking up the content types in the current database
            # before creating foreign keys to them.
            content_type = contenttype_class.objects.db_manager(
                using).get_for_model(cls)
            content_types.add(content_type)
            for params in meta.notifications:
                searched_notifications.append((content_type, params))

    # Find all the Notification that have a content_type for a model we're
    # looking for.  We don't need to check for code names since we already have
    # a list of the ones we're going to create.
    all_notifications = {}
    for item in notification_class.objects.using(using).filter(
            content_type__in=content_types):
        all_notifications[(item.content_type_id, item.codename)] = item

    new_notifications = []
    for ct, params in searched_notifications:
        notification = all_notifications.get((ct.pk, params.codename))
        if notification is None:
            new_notifications.append(params.create(ct))
        else:
            if not params.params_equal(notification):
                if verbosity >= 2:
                    print("Altering notification '%s'" % notification)
                params.set_params(notification)
                if not dry_run:
                    notification.save()
            del all_notifications[(ct.pk, params.codename)]

    # comment for testing
    if verbosity >= 2:
        for notification in new_notifications:
            print("Added notification '%s'" % notification)
        for notification in all_notifications.values():
            print("Removing notification '%s'" % notification)
            if not dry_run:
                notification.delete(using=using)
    if not dry_run:
        notification_class.objects.using(using).bulk_create(new_notifications)



from django.apps import AppConfig
from django.core.management import call_command

try:
    from django.test.utils import setup_databases
except ImportError:  # workaround for django 1.10
    from django.test.runner import setup_databases


class CourierTestConfig(AppConfig):
    name = 'tests'
    verbose_name = 'Courier Test'

    def ready(self):
        setup_databases(verbosity=3, interactive=False)


class CourierDemoConfig(AppConfig):
    name = 'tests'
    verbose_name = 'Courier Demo'

    def ready(self):
        setup_databases(verbosity=3, interactive=False)
        # add notification objects
        from django_courier.management.commands.make_notifications \
            import make_notifications
        make_notifications(self)
        call_command('loaddata', 'demo')



import json
import random
import re

import django.conf
import django.template
import django.template.exceptions
from django.core import mail
from django.db.models import NOT_PROVIDED

from . import settings, base


class MultipartMessage:

    def __init__(self):
        self.subject = ''
        self.text = ''
        self.html = ''

    @classmethod
    def from_dict(cls, obj):
        result = cls()
        result.subject = obj.get('subject')
        result.text = obj.get('text')
        result.html = obj.get('html')
        return result

    @classmethod
    def from_string(cls, string):
        return cls.from_dict(json.loads(string))

    def to_dict(self):
        return {'subject': self.subject, 'text': self.text, 'html': self.html}

    def __str__(self):
        return json.dumps(self.to_dict())

    def to_mail(self) -> mail.EmailMultiAlternatives:
        email = mail.EmailMultiAlternatives()
        email.subject = self.subject
        if self.text:
            email.body = self.text
            if self.html:
                email.attach_alternative(self.html, 'text/html')
        elif self.html:
            email.body = self.html
            email.content_subtype = 'html'
        return email


# from https://stackoverflow.com/questions/2167269/
def from_string(text: str, using=None) -> django.template.base.Template:
    """
    Convert a string into a template object,
    using a given template engine or using the default backends
    from settings.TEMPLATES if no engine was specified.
    """
    # This function is based on django.template.loader.get_template,
    # but uses Engine.from_string instead of Engine.get_template.
    engines = django.template.engines
    engine_list = engines.all() if using is None else [engines[using]]
    exception = None
    for engine in engine_list:
        try:
            return engine.from_string(text).template
        except django.template.exceptions.TemplateSyntaxError as e:
            exception = e
    raise exception


# inspired by django-templated-mail
def email_parts(subject: str, body: str, parameters: dict) -> MultipartMessage:
    message = MultipartMessage()
    html_context = django.template.Context(parameters)
    text_context = django.template.Context(parameters, autoescape=False)
    node_parts = {}
    subject_template = from_string(subject)
    body_template = from_string(body)
    message.subject = subject_template.render(text_context)

    for node in body_template.nodelist:
        name = getattr(node, 'name', None)
        if name is not None:
            node_parts[name] = node

    with html_context.bind_template(body_template):
        with text_context.bind_template(body_template):
            if 'subject' in node_parts:
                message.subject = node_parts['subject'].render(
                    text_context).strip()
            has_parts = False
            if 'text_body' in node_parts:
                has_parts = True
                message.text = node_parts['text_body'].render(
                    text_context).strip()
            if 'html_body' in node_parts:
                has_parts = True
                message.html = node_parts['html_body'].render(
                    html_context).strip()
            if not has_parts:
                message.text = body_template.render(text_context).strip()
    return message


class MarkdownParameters:

    MD_SPECIAL_PATTERN = re.compile(r"[\\\`\*\_\{\}\[\]\(\)\#\+\-\.\!]")

    def __init__(self, obj):
        self._obj = obj

    def __contains__(self, item):
        return self._obj.__contains__(item)

    def __getitem__(self, item):
        obj = self._obj[item]
        if callable(obj):
            return CallableMarkdownParameters(obj)
        return MarkdownParameters(obj)

    def __getattr__(self, attr):
        obj = getattr(self._obj, attr)
        if callable(obj):
            return CallableMarkdownParameters(obj)
        return MarkdownParameters(obj)

    def __str__(self):
        return self.MD_SPECIAL_PATTERN.sub(self.escape, str(self._obj))

    @staticmethod
    def escape(patterns) -> str:
        return '\\' + patterns[0]


class CallableMarkdownParameters(MarkdownParameters):

    def __call__(self, *args, **kwargs):
        obj = self._obj.__call__(*args, **kwargs)
        if callable(obj):
            return CallableMarkdownParameters(obj)
        return MarkdownParameters(obj)


def email_md(subject: str, body: str, parameters: dict) -> MultipartMessage:
    import markdown2
    md_parameters = MarkdownParameters(parameters)
    text_context = django.template.Context(parameters, autoescape=False)
    md_context = django.template.Context(md_parameters, autoescape=False)
    message = MultipartMessage()
    subject_template = from_string(subject)
    body_template = from_string(body)
    message.subject = subject_template.render(text_context)
    message.text = body_template.render(text_context)
    md_body = body_template.render(md_context)
    message.html = markdown2.markdown(
        md_body, extras=settings.MARKDOWN_EXTRAS,
        link_patterns=settings.MARKDOWN_LINK_PATTERNS)
    return message


def email_html(subject: str, body: str, parameters: dict) -> MultipartMessage:
    import august
    text_context = django.template.Context(parameters, autoescape=False)
    html_context = django.template.Context(parameters, autoescape=True)
    message = MultipartMessage()
    # subject
    subject_template = from_string(subject)
    message.subject = subject_template.render(text_context)
    # html
    body_template = from_string(body)
    message.html = body_template.render(html_context)
    message.text = august.convert(message.html)
    return message


def make_model_preview(content_type):
    obj = content_type.objects.first()
    if obj is not None:
        return obj
    obj = content_type()
    for field in content_type._meta.fields:
        value = None
        if field.default != NOT_PROVIDED:
            value = field.default
        if not value:
            desc = str(field.description).lower()
            if desc.startswith('string'):
                value = '{{{}}}'.format(field.verbose_name)
            elif desc.startswith('integer'):
                value = random.randint(1, 100)
            else:
                pass
        setattr(obj, field.attname, value)
    return obj


class PreviewParameters:

    def __init__(self, content_type, sender_model, recipient_model):
        self.recipient = (make_model_preview(recipient_model)
                          if recipient_model else {})
        self.sender = (make_model_preview(sender_model)
                       if sender_model else {})
        self.contact = base.Contact(
            'Contact Name', 'email', 'contact@example.org')
        self.content = make_model_preview(content_type)

    def __contains__(self, item):
        return item in ('contact', 'recipient', 'sender', 'content')

    def __getitem__(self, attr):
        obj = getattr(self, attr)
        return ProxyParameters(obj)


class ProxyParameters:

    def __init__(self, obj):
        self._obj = obj

    def __contains__(self, item):
        return self._obj.__contains__(item)

    def __getitem__(self, item):
        obj = self._obj[item]
        if callable(obj):
            return CallableProxyParameters(obj)
        return ProxyParameters(obj)

    def __getattr__(self, attr):
        obj = getattr(self._obj, attr)
        if callable(obj):
            return CallableProxyParameters(obj)
        return ProxyParameters(obj)

    def __str__(self):
        return str(self._obj)


class CallableProxyParameters(ProxyParameters):
    def __call__(self, *args, **kwargs):
        obj = self._obj.__call__(*args, **kwargs)
        if callable(obj):
            return CallableProxyParameters(obj)
        return ProxyParameters(obj)



import django.http
from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.translation import ugettext_lazy as _

from . import backends, models
from django import forms

# Generic things


class TargetFilter(admin.SimpleListFilter):
    """A really simple filter so that we have the right labels
    """
    title = _('target')
    parameter_name = 'target'

    def lookups(self, request, model_admin):
        """Return list of key/name tuples
        """
        return models.get_target_choices()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(target=self.value())
        return queryset


def target_field(obj):
    return models.get_targets().get(obj.target, obj.target)


class SelectWithSubjectData(forms.Select):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_subjects = {}

    def create_option(self, name, value, label, selected,
                      index, subindex=None, attrs=None):
        index = str(index) if subindex is None else "%s_%s" % (index, subindex)
        option_attrs = {'data-subject': ('true' if self.use_subjects.get(value)
                                         else 'false')}
        if selected:
            option_attrs.update(self.checked_attribute)
        if 'id' in option_attrs:
            option_attrs['id'] = self.id_for_label(option_attrs['id'], index)
        return {
            'name': name,
            'value': value,
            'label': label,
            'selected': selected,
            'index': index,
            'attrs': option_attrs,
            'type': self.input_type,
            'template_name': self.option_template_name,
        }


class BackendChoiceField(forms.ChoiceField):
    widget = SelectWithSubjectData

    def __init__(self, choices=(), *args, **kwargs):
        backs = list(choices)
        choice_pairs = [(back.ID, back.VERBOSE_NAME) for back in backs]
        use_subjects = dict([(back.ID, back.USE_SUBJECT) for back in backs])
        super().__init__(choices=choice_pairs, *args, **kwargs)
        self.widget.use_subjects = use_subjects


class TemplateForm(forms.ModelForm):
    backend = BackendChoiceField(choices=backends.get_backends())
    target = forms.ChoiceField(choices=models.get_target_choices())

    class Meta:
        model = models.Template
        fields = ['notification', 'backend', 'subject', 'content',
                  'target', 'is_active']


class SiteContactForm(forms.ModelForm):
    protocol = forms.ChoiceField(choices=backends.get_protocol_choices())

    class Meta:
        model = models.SiteContact
        fields = ['name', 'protocol', 'address']


class TemplateInline(admin.StackedInline):
    model = models.Template
    form = TemplateForm
    min_num = 0
    extra = 0


class NotificationAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'description', 'use_sender',
                    'use_recipient', 'required', 'template_count')
    list_filter = ('content_type',)
    inlines = [TemplateInline]

    fields = ['codename', 'content_type', 'description',
              'use_sender', 'use_recipient']

    class Media:
        css = {
            'all': ('django_courier/markitup/images.css',
                    'django_courier/markitup/style.css'),
        }
        js = ('django_courier/markitup/jquery.markitup.js',
              'django_courier/notification_fields.js')

    def get_urls(self):
        return [
                   url(
                       r'^(?P<id>\w+)/preview/(?P<backend_id>.+)/$',
                       self.admin_site.admin_view(self.preview),
                       name='django_courier_preview'),
                   url(
                       r'^(?P<id>\w+)/variables/$',
                       self.admin_site.admin_view(self.variables),
                       name='django_courier_variables'),
            ] + super().get_urls()

    def preview(self, request, id, backend_id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        try:
            result = notification.preview(backend_id, request.POST['body'])
        except Exception as exc:
            result = 'Unable to make preview: {}'.format(str(exc))
        return django.http.HttpResponse(result)

    def variables(self, request, id):
        if not self.has_change_permission(request):
            raise PermissionDenied
        notification = self.get_object(request, unquote(id))
        if notification is None:
            raise django.http.Http404(
                _('%(name)s object with primary key %(key)r does not exist.')
                % {'name': force_text(self.model._meta.verbose_name),
                   'key': escape(id)})
        if request.method != 'POST':
            return django.http.HttpResponseNotAllowed(('POST',))
        result = notification.get_variables()
        return django.http.JsonResponse(result, safe=False)

    def get_readonly_fields(self, request, obj=None):
        return ['codename', 'content_type', 'description',
                'use_sender', 'use_recipient', 'required',
                'sender_model', 'recipient_model']

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    @staticmethod
    def template_count(obj):
        return obj.template_set.count()


class SiteContactPreferenceInline(admin.TabularInline):
    model = models.SiteContactPreference


class SiteContactAdmin(admin.ModelAdmin):
    form = SiteContactForm
    inlines = [
        SiteContactPreferenceInline
    ]
    list_display = ('name', 'address', 'protocol')


class FailedMessageAdmin(admin.ModelAdmin):
    list_display = ('backend', 'address', 'created_at')
    list_filter = ('backend', 'address')


admin.site.register(models.SiteContact, SiteContactAdmin)
admin.site.register(models.FailedMessage, FailedMessageAdmin)
admin.site.register(models.Notification, NotificationAdmin)



"""
Django settings for django_courier tests
"""
# based of django-debug-toolbar
# https://github.com/jazzband/django-debug-toolbar/blob/master/tests/settings.py
import os

from django.utils.translation import ugettext_lazy as _

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production

SECRET_KEY = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'

INTERNAL_IPS = ['127.0.0.1']
ALLOWED_HOSTS = ['127.0.0.1']

LOGGING_CONFIG = None   # avoids spurious output in tests
DEBUG = True

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django_courier',
    'tests',
]

AUTH_USER_MODEL = 'tests.User'

FIXTURE_DIRS = ['fixtures']

MEDIA_URL = '/media/'   # Avoids https://code.djangoproject.com/ticket/21451

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.root_urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

STATIC_ROOT = os.path.join(BASE_DIR, 'tests', 'static')

STATIC_URL = '/static/'

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, 'tests', 'additional_static'),
    ("prefix", os.path.join(BASE_DIR, 'tests', 'additional_static')),
]

# Cache and database

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
    'second': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
}

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

DJANGO_COURIER_CHANNELS = {
    '': (_('Sender'), _('Recipient')),
    'sub': (_('Sender\'s Subscribers'), _('Recipient\'s Subscribers')),
}
DJANGO_COURIER_SENDER_MODEL = AUTH_USER_MODEL
DJANGO_COURIER_RECIPIENT_MODEL = AUTH_USER_MODEL



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-02-08 13:32
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0002_site_contacts'),
    ]

    operations = [
        migrations.CreateModel(
            name='FailedMessage',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False,
                    verbose_name='ID')),
                ('backend', models.CharField(max_length=100)),
                ('address', models.CharField(max_length=500)),
                ('protocol', models.CharField(max_length=100)),
                ('message', models.TextField()),
                ('error', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'verbose_name': 'failed message',
                'default_permissions': (),
            },
        ),
        migrations.AlterField(
            model_name='template',
            name='send_to_site_contacts',
            field=models.BooleanField(
                default=False,
                help_text='Whether this message is sent to the site contacts '
                          'or to the notification recipient.'),
        ),
    ]



import re

import django.conf
from django.utils.translation import ugettext_lazy as _

URL_PATTERN = re.compile(
    r'((([A-Za-z]{3,9}:(?:\/\/)?)(?:[\-;:&=\+\$,\w]+@)?[A-Za-z0-9\.\-]'
    r'+(:[0-9]+)?|(?:www\.|[\-;:&=\+\$,\w]+@)[A-Za-z0-9\.\-]+)((?:\/[\+'
    r'~%\/\.\w\-_]*)?\??(?:[\-\+=&;%@\.\w_]*)#?(?:[\.\!\/\\\w]*))?)')

BACKENDS = getattr(django.conf.settings, 'DJANGO_COURIER_BACKENDS', None)
if BACKENDS is None:
    # Automatically set based on libraries available
    BACKENDS = ['django_courier.backends.EmailBackend',
                'django_courier.backends.SlackWebhookBackend']
    try:
        import august  # noqa: F401
        BACKENDS.append('django_courier.backends.HtmlEmailBackend')
    except ImportError:
        pass
    try:
        import markdown2  # noqa: F401
        BACKENDS.append('django_courier.backends.MarkdownEmailBackend')
    except ImportError:
        pass
    try:
        import twilio.rest  # noqa: F401
        BACKENDS.append('django_courier.backends.TwilioBackend')
    except ImportError:
        pass

CHANNELS = getattr(django.conf.settings, 'DJANGO_COURIER_CHANNELS', None)
if CHANNELS is None:
    CHANNELS = {
        '': (_('Sender'), _('Recipient')),
    }

MARKDOWN_EXTRAS = getattr(django.conf.settings,
                          'DJANGO_COURIER_MARKDOWN_EXTRAS', None)
if MARKDOWN_EXTRAS is None:
    MARKDOWN_EXTRAS = ['footnotes', 'link_patterns', 'smarty-pants', 'tables']

MARKDOWN_LINK_PATTERNS = getattr(django.conf.settings,
                                 'DJANGO_COURIER_MARKDOWN_LINK_PATTERNS', None)
if MARKDOWN_LINK_PATTERNS is None:
    MARKDOWN_LINK_PATTERNS = [(URL_PATTERN, r'\1')]

SENDER_MODEL = getattr(django.conf.settings,
                       'DJANGO_COURIER_SENDER_MODEL', None)
RECIPIENT_MODEL = getattr(django.conf.settings,
                          'DJANGO_COURIER_RECIPIENT_MODEL', None)



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_courier', '0009_subjects'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='recipient_model',
            field=models.CharField(
                blank=True, max_length=500, verbose_name='recipient model'),
        ),
        migrations.AddField(
            model_name='notification',
            name='sender_model',
            field=models.CharField(
                blank=True, max_length=500, verbose_name='sender model'),
        ),
    ]



from django.contrib import admin

from . import models

admin.site.register(models.User)
admin.site.register(models.Article)
admin.site.register(models.Subscriber)
admin.site.register(models.Comment)



from django.shortcuts import get_object_or_404, redirect, render

from . import models


def article_list(request):
    context = {
        'articles': models.Article.objects.order_by('-id'),
    }
    return render(request, 'tests/index.html', context)


def article_detail(request, article_id):
    article = get_object_or_404(models.Article, pk=article_id)
    comments = models.Comment.objects.filter(article=article).order_by('id')
    token = request.GET.get('token')
    return render(request, 'tests/detail.html', {
        'article': article, 'comments': comments, 'token': token})


def comment(request, article_id):
    token = request.POST.get('token', '')
    content = request.POST.get('content', '')
    article = get_object_or_404(models.Article, pk=article_id)
    subscriber = models.Subscriber.load_from_token(token)
    models.Comment.objects.create(
        content=content,
        poster=subscriber,
        article=article,
    )
    return redirect('tests:detail', article.id)


def subscribe(request):
    name = request.POST.get('name', '')
    email = request.POST.get('email', '')
    if email == '' or name == '':
        return render(request, 'tests/index.html', {
            'error_message': 'You are missing an email or a name', })

    models.Subscriber.objects.create(name=name, email=email)
    return render(request, 'tests/subscribed.html')



import abc
from typing import Any, Iterable, List, TypeVar


class Contact:
    """A generic contact object

    If you want to return something that looks like this, make sure
    to implement the __hash__ method the same, otherwise filtering
    duplicate contacts won't work
    """

    def __init__(self, name: str, protocol: str, address: str, obj: Any=None):
        self.name = name
        self.protocol = protocol
        self.address = address
        self.object = obj

    def __str__(self):
        return '{} <{}:{}>'.format(self.name, self.protocol, self.address)

    def __hash__(self):
        return hash(str(self))


class AbstractContactable(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_contacts_for_notification(
            self, notification: 'Notification') -> List[Contact]:
        ...


class AbstractContactNetwork(metaclass=abc.ABCMeta):

    def get_contactables(self, channel: str) -> List[AbstractContactable]:
        ...


class ContactNetwork:

    def get_contactables(self, channel: str) -> Iterable[AbstractContactable]:
        if channel == '':
            return (self,)
        raise ValueError('Channel {} not supported'.format(channel))


AbstractContactable.register(ContactNetwork)
AbstractContactNetwork.register(ContactNetwork)


AbstractContactNetworkN = TypeVar('AbstractContactNetworkN',
                                  AbstractContactNetwork, None)
